Friend Class ReflectionCache

    Private Shared _Instances As Dictionary(Of Type, ReflectionCache) = New Dictionary(Of Type, ReflectionCache)

    Public Shared Function CreateInstance(ByVal ForType As Type) As ReflectionCache

        If Not _Instances.ContainsKey(ForType) Then
            Try
                _Instances.Add(ForType, New ReflectionCache(ForType))
            Catch ex As ArgumentException
                '
                '   This actually can happen in multithreaded scenarios: Thread A queries
                '   ContainsKey; Thread B queries ContainsKey; Thread A adds new item;
                '   Thread B adds new item.
                '
            End Try
        End If
        Return _Instances(ForType)

    End Function

    Private _ColumnAttributes As Dictionary(Of String, ColumnAttribute)
    Private _ColumnMap As Dictionary(Of String, String)
    Private _IsRowObject As Boolean
    Private _Members As Dictionary(Of String, MemberInfo)
    Private _ParamAttributes As Dictionary(Of String, ParameterAttribute)
    Private _ReturnValueMember As String
    Private _StoredProcedureAttrib As StoredProcedureAttribute
    Private _T As Type

    Private Sub New(ByVal ForType As Type)

        _T = ForType
        _Members = New Dictionary(Of String, MemberInfo)
        _ParamAttributes = New Dictionary(Of String, ParameterAttribute)
        _ColumnAttributes = New Dictionary(Of String, ColumnAttribute)
        _ColumnMap = New Dictionary(Of String, String)

        Dim Attribs As Object() = _T.GetCustomAttributes(GetType(StoredProcedureAttribute), True)
        If Attribs.Length = 1 Then
            _StoredProcedureAttrib = CType(Attribs(0), StoredProcedureAttribute)
        End If

        Dim RelevantMembers As New List(Of MemberInfo)
        RelevantMembers.AddRange(_T.GetFields(BindingFlags.Instance Or BindingFlags.Public))
        RelevantMembers.AddRange(_T.GetProperties(BindingFlags.Instance Or BindingFlags.Public))
        For Each Member As MemberInfo In RelevantMembers
            For Each Attribute As Object In Member.GetCustomAttributes(True)
                If TypeOf Attribute Is ParameterAttribute Then
                    _Members.Add(Member.Name, Member)
                    _ParamAttributes.Add(Member.Name, CType(Attribute, ParameterAttribute))
                ElseIf TypeOf Attribute Is ReturnValueAttribute Then
                    _Members.Add(Member.Name, Member)
                    _ReturnValueMember = Member.Name
                ElseIf TypeOf Attribute Is ColumnAttribute Then
                    _Members.Add(Member.Name, Member)
                    If String.IsNullOrEmpty(CType(Attribute, ColumnAttribute).ColumnName) Then
                        _ColumnMap.Add(Member.Name, Member.Name)
                        _ColumnAttributes.Add(Member.Name, CType(Attribute, ColumnAttribute))
                    Else
                        _ColumnMap.Add(CType(Attribute, ColumnAttribute).ColumnName, Member.Name)
                        _ColumnAttributes.Add(Member.Name, CType(Attribute, ColumnAttribute))
                    End If
                ElseIf TypeOf Attribute Is RowAttribute Then
                    _IsRowObject = True
                End If
            Next
        Next

    End Sub

    Public Function GetColumnAttribute(ByVal ForMember As String) As ColumnAttribute

        Return _ColumnAttributes(ForMember)

    End Function

    Public Function GetColumnMap() As Dictionary(Of String, String)

        Return _ColumnMap

    End Function

    Public Function GetMember(ByVal Name As String) As MemberInfo

        Return _Members(Name)

    End Function

    Public Function GetMemberNames() As List(Of String)
        Dim Names As List(Of String)

        Names = New List(Of String)(_Members.Keys)
        Return Names

    End Function

    Public Function GetMemberType(ByVal MemberName As String) As Type
        Dim Member As MemberInfo

        Member = GetMember(MemberName)
        If TypeOf Member Is FieldInfo Then
            Return CType(Member, FieldInfo).FieldType
        Else
            Return CType(Member, PropertyInfo).PropertyType
        End If

    End Function

    Public Function GetParameterAttribute(ByVal MemberName As String) As ParameterAttribute

        If _ParamAttributes.ContainsKey(MemberName) Then
            Return _ParamAttributes(MemberName)
        Else
            Return Nothing
        End If

    End Function

    Public Function GetReturnValueMember() As String

        Return _ReturnValueMember

    End Function

    Public Function GetStoredProcedureAttribute() As StoredProcedureAttribute

        Return _StoredProcedureAttrib

    End Function

    Public ReadOnly Property IsRowObject() As Boolean
        Get
            Return _IsRowObject
        End Get
    End Property

End Class
