Imports System.Data.Common
Imports System.Data.SqlClient
Imports System.Configuration

''' <summary>
''' This class implementats IDataService for MS SQL Server databases.
''' </summary>
Public Class SqlServerDataService
    Implements IDataService

    Private _Connection As IDbConnection
    Private _ConnectionName As String
    Private _ConnectionString As String
    Private _Factory As DbProviderFactory
    Private _InTransaction As Boolean
    Private _Transaction As IDbTransaction
    Private _TypeMap As Dictionary(Of Type, DbType)

    Public Sub New()

        InitializeTypeMap()

    End Sub

    Private Sub Abort() Implements IDataService.Abort

        If _InTransaction Then
            _Transaction.Rollback()
            _Connection.Close()
            _Connection = Nothing
            _InTransaction = False
        Else
            Throw New InvalidOperationException("Not in a transaction")
        End If

    End Sub

    Private Sub BeginTransaction() Implements IDataService.BeginTransaction

        If Not _InTransaction Then
            If _Connection Is Nothing OrElse _Connection.State <> ConnectionState.Open Then
                _Connection = CreateConnection()
            End If
            _Transaction = _Connection.BeginTransaction
            _InTransaction = True
        End If

    End Sub

    Private Sub Commit() Implements IDataService.Commit

        If _InTransaction Then
            _Transaction.Commit()
            _Connection.Close()
            _Connection = Nothing
            _InTransaction = False
        Else
            Throw New InvalidOperationException("Not in a transaction")
        End If

    End Sub

    Private Property ConnectionName() As String Implements IDataService.ConnectionName
        Get
            Return _ConnectionName
        End Get
        Set(ByVal value As String)
            Dim ConnectInfo As ConnectionStringSettings

            If String.IsNullOrEmpty(_ConnectionName) Then
                _ConnectionName = value
                ConnectInfo = ConfigurationManager.ConnectionStrings(_ConnectionName)
                _Factory = DbProviderFactories.GetFactory(ConnectInfo.ProviderName)
                _ConnectionString = ConnectInfo.ConnectionString
            Else
                Throw New InvalidOperationException("Property is read-only at this time.")
            End If

        End Set
    End Property

    Private Function CreateCommand() As IDbCommand
        Dim Cmd As IDbCommand

        Cmd = _Factory.CreateCommand
        If _InTransaction Then
            Cmd.Transaction = _Transaction
        End If
        Return Cmd

    End Function

    Private Function CreateCommand(ByVal CallObject As CallObject) As IDbCommand
        Dim Cmd As IDbCommand
        Dim StoredProcAttrib As StoredProcedureAttribute
        Dim Reflect As ReflectionCache
        Dim p As IDbDataParameter
        Dim ParamAttrib As ParameterAttribute
        Dim MemberType As Type

        Reflect = ReflectionCache.CreateInstance(CType(CallObject, Object).GetType)
        Cmd = CreateCommand()
        Cmd.CommandType = CommandType.StoredProcedure
        StoredProcAttrib = Reflect.GetStoredProcedureAttribute
        If String.IsNullOrEmpty(StoredProcAttrib.Name) Then
            Cmd.CommandText = CType(CallObject, Object).GetType.Name
        Else
            Cmd.CommandText = StoredProcAttrib.Name
        End If
        p = Cmd.CreateParameter
        p.Direction = ParameterDirection.ReturnValue
        p.ParameterName = "ReturnValue"
        Cmd.Parameters.Add(p)
        For Each MemberName As String In Reflect.GetMemberNames()
            ParamAttrib = Reflect.GetParameterAttribute(MemberName)
            If ParamAttrib IsNot Nothing Then
                p = New SqlParameter
                If Convert.ToInt32(ParamAttrib.Type) = -1 Then
                    MemberType = Reflect.GetMemberType(MemberName)
                    If _TypeMap.ContainsKey(MemberType) Then
                        p.DbType = _TypeMap(MemberType)
                    Else
                        '
                        '   We don't have a TypeMap entry for the type of this member, so we
                        '   take our chances letting the SqlClient figure out what type this
                        '   parameter is
                        '
                    End If
                Else
                    '
                    '   The attribute has specified a type for this member, so use it.
                    '
                    p.DbType = ParamAttrib.Type
                End If
                p.Direction = ParamAttrib.Direction
                If String.IsNullOrEmpty(ParamAttrib.Name) Then
                    p.ParameterName = FormatParameterName(MemberName)
                Else
                    p.ParameterName = FormatParameterName(ParamAttrib.Name)
                End If
                If ParamAttrib.Size <> -1 Then
                    p.Size = ParamAttrib.Size
                End If
                If ParamAttrib.Direction = ParameterDirection.Input Or ParamAttrib.Direction = ParameterDirection.InputOutput Then
                    p.Value = CallObject.GetProperty(MemberName, Nothing)
                    If p.Value Is Nothing Then
                        p.Value = DBNull.Value
                    End If
                End If
                Cmd.Parameters.Add(p)
            End If
        Next
        Return Cmd

    End Function

    Private Function CreateConnection() As IDbConnection
        Dim DB As IDbConnection

        DB = _Factory.CreateConnection
        DB.ConnectionString = _ConnectionString
        DB.Open()
        Return DB

    End Function

    Private Sub Execute(ByVal CallObject As CallObject) Implements IDataService.Execute

        If _InTransaction Then
            Execute(CallObject, _Connection, _Transaction)
        Else
            Dim DB As IDbConnection
            Dim TX As IDbTransaction

            DB = CreateConnection()
            TX = DB.BeginTransaction
            Try
                Execute(CallObject, DB, TX)
                TX.Commit()
            Catch ex As Exception
                TX.Rollback()
                Throw
            Finally
                If DB IsNot Nothing Then
                    DB.Close()
                End If
            End Try
        End If

    End Sub

    Private Sub Execute(ByVal CallObject As CallObject, ByVal Connection As IDbConnection, ByVal Transaction As IDbTransaction)
        Dim Cmd As IDbCommand
        Dim Reflect As ReflectionCache
        Dim ParamAttrib As ParameterAttribute
        Dim ParamName As String
        Dim Value As Object
        Dim ReturnValueMember As String

        Cmd = CreateCommand(CallObject)
        Cmd.Transaction = Transaction
        Cmd.Connection = Connection
        Cmd.ExecuteNonQuery()
        Reflect = ReflectionCache.CreateInstance(CType(CallObject, Object).GetType)
        For Each MemberName As String In Reflect.GetMemberNames()
            ParamAttrib = Reflect.GetParameterAttribute(MemberName)
            If ParamAttrib IsNot Nothing AndAlso (ParamAttrib.Direction = ParameterDirection.InputOutput Or ParamAttrib.Direction = ParameterDirection.Output) Then
                If String.IsNullOrEmpty(ParamAttrib.Name) Then
                    ParamName = FormatParameterName(MemberName)
                Else
                    ParamName = FormatParameterName(ParamAttrib.Name)
                End If
                Value = CType(Cmd.Parameters(ParamName), IDbDataParameter).Value
                If Value IsNot DBNull.Value Then
                    CallObject.SetProperty(MemberName, Value)
                End If
            End If
        Next
        ReturnValueMember = Reflect.GetReturnValueMember()
        If Not String.IsNullOrEmpty(ReturnValueMember) Then
            CallObject.SetProperty(ReturnValueMember, CType(Cmd.Parameters("ReturnValue"), IDbDataParameter).Value)
        End If

    End Sub

    Private Function FormatParameterName(ByVal s As String) As String

        If s.Chars(0) = "@"c Then
            Return s
        Else
            Return String.Concat("@", s)
        End If

    End Function

    Private Sub InitializeTypeMap()

        _TypeMap = New Dictionary(Of Type, DbType)
        _TypeMap.Add(GetType(Int16), DbType.Int16)
        _TypeMap.Add(GetType(Nullable(Of Int16)), DbType.Int16)
        _TypeMap.Add(GetType(Int32), DbType.Int32)
        _TypeMap.Add(GetType(Nullable(Of Int32)), DbType.Int32)
        _TypeMap.Add(GetType(Int64), DbType.Int64)
        _TypeMap.Add(GetType(Nullable(Of Int64)), DbType.Int64)
        _TypeMap.Add(GetType(String), DbType.String)
        _TypeMap.Add(GetType(DateTime), DbType.DateTime)
        _TypeMap.Add(GetType(Nullable(Of DateTime)), DbType.DateTime)
        _TypeMap.Add(GetType(Byte), DbType.Byte)
        _TypeMap.Add(GetType(Nullable(Of Byte)), DbType.Byte)
        _TypeMap.Add(GetType(Byte()), DbType.Binary)
        _TypeMap.Add(GetType(Boolean), DbType.Boolean)
        _TypeMap.Add(GetType(Nullable(Of Boolean)), DbType.Boolean)
        _TypeMap.Add(GetType(Guid), DbType.Guid)
        _TypeMap.Add(GetType(Nullable(Of Guid)), DbType.Guid)
        _TypeMap.Add(GetType(Decimal), DbType.Decimal)
        _TypeMap.Add(GetType(Nullable(Of Decimal)), DbType.Decimal)
        _TypeMap.Add(GetType(Double), DbType.Double)
        _TypeMap.Add(GetType(Nullable(Of Double)), DbType.Double)
        _TypeMap.Add(GetType(Single), DbType.Single)
        _TypeMap.Add(GetType(Nullable(Of Single)), DbType.Single)
        _TypeMap.Add(GetType([Enum]), DbType.Int32)

    End Sub

    Private Function Query(Of T As RowObject)(ByVal CallObject As CallObject) As System.Collections.Generic.List(Of T) Implements IDataService.Query

        If _InTransaction Then
            Return Query(Of T)(CallObject, _Connection, _Transaction)
        Else
            Dim DB As IDbConnection
            Dim TX As IDbTransaction

            DB = CreateConnection()
            TX = DB.BeginTransaction
            Try
                Return Query(Of T)(CallObject, DB, TX)
                TX.Commit()
            Catch ex As Exception
                TX.Rollback()
                Throw
            Finally
                If DB IsNot Nothing Then
                    DB.Close()
                End If
            End Try
        End If

    End Function

    Private Function Query(Of T As RowObject)(ByVal CallObject As CallObject, ByVal Connection As IDbConnection, ByVal Transaction As IDbTransaction) As System.Collections.Generic.List(Of T)
        Dim Collection As List(Of T)
        Dim Reflect As ReflectionCache
        Dim ColumnMap As Dictionary(Of String, String)
        Dim Cmd As IDbCommand = Nothing
        Dim ResultsetColumn As String
        Dim CurrentRow As T
        Dim MemberName As String
        Dim Value As Object
        Dim Attrib As ColumnAttribute

        Collection = New List(Of T)
        Reflect = ReflectionCache.CreateInstance(GetType(T))
        ColumnMap = Reflect.GetColumnMap
        Cmd = CreateCommand(CallObject)
        Cmd.Connection = Connection
        Cmd.Transaction = Transaction
        Using Results As IDataReader = Cmd.ExecuteReader()
            While Results.Read()
                CurrentRow = Activator.CreateInstance(Of T)()
                For Each ResultsetColumn In ColumnMap.Keys
                    MemberName = Reflect.GetColumnMap(ResultsetColumn)
                    Value = Results(ResultsetColumn)
                    If Value Is DBNull.Value Then
                        Attrib = Reflect.GetColumnAttribute(MemberName)
                        If Attrib.NullValue IsNot Nothing Then
                            CurrentRow.SetProperty(MemberName, Attrib.NullValue)
                        End If
                    Else
                        CurrentRow.SetProperty(MemberName, Value)
                    End If
                Next
                Collection.Add(CurrentRow)
            End While
        End Using
        Return Collection

    End Function

End Class


